
#include <gtk/gtk.h>
#include <hildon/hildon.h>
#include <dbus/dbus-glib.h>

#include <libintl.h>
#include <locale.h>
#include <string.h>

#define CONFIG_PATH "/usr/share/applications/hildon/"
#define GETTEXT_PACKAGE "maemo-af-desktop"
#define APP_MGR_DBUS_NAME "com.nokia.HildonDesktop.AppMgr"
#define APP_MGR_DBUS_PATH "/com/nokia/HildonDesktop/AppMgr"
#define APP_MGR_LAUNCH_METHODE "LaunchApplication"
#define DESKTOP_ENTRY_CONFIG_GROUP "Desktop Entry"
#define TEXTDOMAIN_KEY "X-Text-Domain"
#define NAME_KEY "Name"
#define COMMENT_KEY "Comment"
#define ICON_KEY "Icon"
#define NODISPLAY_KEY "NoDisplay"

struct _AppSearch
{
  GtkListStore* app_list;
  GtkTreeModel* filtered_app_list;
  GtkTreeView* tree_view;
  HildonPannableArea* pannable_area;
  GtkWidget* entry;
  GtkWidget *dialog;
};

typedef struct _AppSearch AppSearch;

typedef struct _result_data_t
{
  AppSearch *app_search;
  gchar* icon_name;
  gchar* app_display_name;
  gchar* app_name;
  gchar* app_comment;
  gchar* desktop_file_name;
} result_data_t;


const gchar*
_get_default_text_color()
{
  static gchar buf[40] = {0};
  if(buf[0] == '\0')
  {
    GdkColor color;
    GtkStyle *style = gtk_rc_get_style_by_paths(gtk_settings_get_default(),
						NULL, NULL, GTK_TYPE_LABEL);
    if(gtk_style_lookup_color(style, "DefaultTextColor", &color))
      sprintf(buf, "#%02x%02x%02x", color.red / 256, color.green / 256, color.blue / 256);
  }
  return buf;
}

const gchar*
_get_active_text_color()
{
  static gchar buf[40] = {0};
  if(buf[0] == '\0')
  {
    GdkColor color;
    GtkStyle *style = gtk_rc_get_style_by_paths(gtk_settings_get_default(),
						NULL, NULL, GTK_TYPE_LABEL);
    if(gtk_style_lookup_color(style, "SecondaryTextColor", &color))
      sprintf(buf, "#%02x%02x%02x", color.red / 256, color.green / 256, color.blue / 256);
  }
  return buf;
}

static
gchar* 
_retrieve_app_name(GKeyFile* key_file)
{
  if(g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, TEXTDOMAIN_KEY, NULL) &&
     g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, NAME_KEY, NULL))
  {
    gchar* text_domain = g_key_file_get_string(key_file, DESKTOP_ENTRY_CONFIG_GROUP, TEXTDOMAIN_KEY, NULL);
    gchar* app_name_key = g_key_file_get_string(key_file, DESKTOP_ENTRY_CONFIG_GROUP, NAME_KEY, NULL);
    gchar* app_name = g_strdup(dgettext(text_domain, app_name_key));
    g_free(text_domain);
    g_free(app_name_key);
    
    return app_name;
  }
  else if(g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, NAME_KEY, NULL))
  {
    gchar* app_name = g_key_file_get_string(key_file, DESKTOP_ENTRY_CONFIG_GROUP, NAME_KEY, NULL);
    gchar* name = g_strdup(dgettext(GETTEXT_PACKAGE, app_name));
    g_free(app_name);
    return name;
  }
  return NULL;
}

static
gchar* 
_retrieve_app_comment(GKeyFile* key_file)
{
  if(g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, TEXTDOMAIN_KEY, NULL) &&
     g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, COMMENT_KEY, NULL))
  {
    gchar* text_domain = g_key_file_get_string(key_file, DESKTOP_ENTRY_CONFIG_GROUP, TEXTDOMAIN_KEY, NULL);
    gchar* app_comment_key = g_key_file_get_string(key_file, DESKTOP_ENTRY_CONFIG_GROUP, COMMENT_KEY, NULL);
    gchar* app_comment = g_strdup(dgettext(text_domain, app_comment_key));
    g_free(text_domain);
    if(g_strcmp0(app_comment, app_comment_key) == 0)
    {
      g_free(app_comment);
      app_comment = NULL;
    }
    g_free(app_comment_key);
    return app_comment;
  }
  else if(g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, COMMENT_KEY, NULL))
  {
    gchar* app_comment = g_key_file_get_string(key_file, DESKTOP_ENTRY_CONFIG_GROUP, COMMENT_KEY, NULL);
    gchar* comment = g_strdup(dgettext(GETTEXT_PACKAGE, app_comment));
    g_free(app_comment);
    return comment;
  }
  return NULL;
}


static gchar*
_retrieve_icon_name(GKeyFile* key_file)
{
  if(g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, ICON_KEY, NULL)) {
      return g_key_file_get_string(key_file, DESKTOP_ENTRY_CONFIG_GROUP, ICON_KEY, NULL);
  } else {
      return NULL;
  }
}

static gboolean
_shouldnt_be_visible(GKeyFile* key_file)
{
  if(g_key_file_has_key(key_file, DESKTOP_ENTRY_CONFIG_GROUP, NODISPLAY_KEY, NULL))
  {
    gboolean no_display = g_key_file_get_boolean(key_file, DESKTOP_ENTRY_CONFIG_GROUP, NODISPLAY_KEY, NULL);
    return no_display;
  }
  return FALSE;
}

static gboolean
_append_result_entry(gpointer user_data)
{
  result_data_t* result_data = (result_data_t*)user_data;
  GtkTreeIter iter;
  gtk_list_store_append(result_data->app_search->app_list, &iter);
  gtk_list_store_set(result_data->app_search->app_list, &iter,
		     0, result_data->icon_name,
		     1, result_data->app_display_name,
		     2, result_data->app_name,
		     3, result_data->app_comment,
		     4, result_data->desktop_file_name,
		     -1);
  g_free(result_data->app_display_name);
  g_free(result_data->app_name);
  g_free(result_data->app_comment);
  g_free(result_data->desktop_file_name);
  g_free(result_data);
  return FALSE;
}

static void
_add_app_from_desktop_file_to_liststore(const gchar* file_name, AppSearch* app_search)
{
  GKeyFile *key_file = g_key_file_new();
  if(g_key_file_load_from_file(key_file, file_name, G_KEY_FILE_KEEP_COMMENTS, NULL))
  {
    if(_shouldnt_be_visible(key_file))
    {
    }
    else
    {
      gchar* app_name = _retrieve_app_name(key_file);

      if(app_name == NULL)
      {
	return;
      }
      gchar* app_comment = _retrieve_app_comment(key_file);
      gchar* icon = _retrieve_icon_name(key_file);
      gchar* desktop_file_name = g_path_get_basename(file_name);
      result_data_t* result_data = g_new0(result_data_t, 1);
      result_data->app_search = app_search;
      result_data->icon_name = icon;
      result_data->app_name = g_strdup(app_name);
      if(app_comment == NULL)
      {
	result_data->app_display_name = 
	  g_markup_printf_escaped("<span color=\"%s\">%s</span>", _get_default_text_color(), app_name);
	result_data->app_comment = g_strdup("");
      }
      else
      {
	result_data->app_comment = g_strdup(app_comment);
	result_data->app_display_name = 
	  g_markup_printf_escaped("<span color=\"%s\">%s</span>\n"
				  "<span size=\"x-small\" color=\"%s\">%s</span>", 
				  _get_default_text_color(), app_name,
				  _get_active_text_color(), app_comment);
      }
      result_data->desktop_file_name = g_strdup(desktop_file_name);
      g_idle_add((GSourceFunc)_append_result_entry, result_data);
      g_free(desktop_file_name);
      g_free(app_name);
      g_free(app_comment);
      //      g_object_unref(icon);
    }
    g_key_file_free(key_file);
  }
}

static void
_app_list_init_func(gpointer user_data)
{
  AppSearch *app_search = (AppSearch*)user_data;
  GError* error = NULL;

  GDir* directory = g_dir_open(CONFIG_PATH, 0, &error);
  if(error)
  {
    g_warning("error on open directory %s\n", CONFIG_PATH);
    g_error_free(error);
    error = NULL;
  }
  else
  {
    const gchar* file_name = NULL;
    while((file_name = g_dir_read_name(directory))!=NULL)
    {
      gchar* full_path = g_build_filename(CONFIG_PATH, file_name, NULL);
      _add_app_from_desktop_file_to_liststore(full_path, app_search);
      g_free(full_path);
    }
    g_dir_close(directory);
  }
  return;
}

static void
_call_application(AppSearch *app_search)
{
  DBusGConnection *connection;
  GError *error = NULL;
  connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
  if(error)
  {
    g_warning ("Can not connect to session bus. %s",error->message);
    g_error_free (error);
    return ;
  }
  DBusGProxy* proxy = NULL;
  proxy = dbus_g_proxy_new_for_name(connection,
				    APP_MGR_DBUS_NAME,
				    APP_MGR_DBUS_PATH,
				    APP_MGR_DBUS_NAME);

  if(!proxy)
  {
    g_warning ("Can not connect dbus proxy.");
    return;
  }

  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;

  selection = gtk_tree_view_get_selection(app_search->tree_view);

  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    gchar* app_desktop_file = NULL;
    gtk_tree_model_get(model,
		       &iter,
		       4, &app_desktop_file,
		       -1);
    if(app_desktop_file && g_str_has_suffix(app_desktop_file, ".desktop"))
    {
      //gtk_widget_hide(app_search->dialog);
      gchar *id = g_strndup(app_desktop_file, strlen (app_desktop_file) - strlen (".desktop"));
      dbus_g_proxy_call_no_reply (proxy,
                                  APP_MGR_LAUNCH_METHODE,
                                  G_TYPE_STRING,
                                  id,
                                  G_TYPE_INVALID,
                                  G_TYPE_INVALID);
      g_free (id);
    }
    if(app_desktop_file)
      g_free(app_desktop_file);
  }
  g_object_unref(proxy);
}

static gboolean
_filter_app_list_func(GtkTreeModel* model,
		      GtkTreeIter* iter,
		      gpointer data)
{
  
  AppSearch *app_search = (AppSearch*)data;
  GtkEntry* entry = GTK_ENTRY(app_search->entry);
  
  if(gtk_entry_get_text_length(entry)>0)
  {
    gchar* text = g_utf8_strdown(gtk_entry_get_text(entry),-1);
    gchar* app_name = NULL;
    gchar* app_comment = NULL;
    gtk_tree_model_get(model, iter,
		       2, &app_name,
		       3, &app_comment,
		       -1);
    gboolean visible = FALSE;
    if(app_name)
    {
      gchar* app_name_downcase = g_utf8_strdown(app_name, -1);
      gchar* app_comment_downcase = g_utf8_strdown(app_comment, -1);
      if(strstr(app_name_downcase, text)!=NULL ||
	 strstr(app_comment_downcase, text)!=NULL)
      {
	visible = TRUE;
      }
      g_free(app_name);
      g_free(app_name_downcase);
      g_free(app_comment);
      g_free(app_comment_downcase);
    }
    g_free(text);
    return visible;
  }
  
  return TRUE;
}

static void
_on_activate(GtkWidget *widget, gpointer user_data)
{
    AppSearch *app_search = (AppSearch*)user_data;
    _call_application(app_search);
    gtk_dialog_response(GTK_DIALOG(app_search->dialog), GTK_RESPONSE_OK);
}

static gboolean
_on_key_press(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
    AppSearch *app_search = (AppSearch*)user_data;
    GtkTreeSelection *selection = gtk_tree_view_get_selection(app_search->tree_view);
    GtkTreeModel *model;
    GtkTreeIter iter;
    gboolean result = FALSE;

    switch (event->keyval) {
        case GDK_BackSpace:
            /* Close the dialog when backspacing on empty widget */
            if (widget == GTK_WIDGET(app_search->entry) &&
                    gtk_entry_get_text_length(GTK_ENTRY(app_search->entry)) == 0) {
                gtk_dialog_response(GTK_DIALOG(app_search->dialog), GTK_RESPONSE_OK);
            }
            break;
        case GDK_Up:
        case GDK_Left:
            if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
                GtkTreePath *path;
                path = gtk_tree_model_get_path(model, &iter);
                if (path && gtk_tree_path_prev(path)) {
                    if (gtk_tree_model_get_iter(model, &iter, path)) {
                        gtk_tree_selection_select_iter(selection, &iter);
                    }
                }
                gtk_tree_path_free(path);
            } else {
                if (gtk_tree_model_get_iter_first(model, &iter)) {
                    gtk_tree_selection_select_iter(selection, &iter);
                }
            }
            result = TRUE;
            break;
        case GDK_Down:
        case GDK_Right:
            if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
                if (gtk_tree_model_iter_next(model, &iter)) {
                    gtk_tree_selection_select_iter(selection, &iter);
                }
            } else {
                if (gtk_tree_model_get_iter_first(model, &iter)) {
                    gtk_tree_selection_select_iter(selection, &iter);
                }
            }
            result = TRUE;
            break;
        default:
            break;
    }

    if (result) {
        GtkTreePath *path;
        GdkRectangle rect;
        gint y;

        path = gtk_tree_model_get_path (model, &iter);
        gtk_tree_view_get_background_area (GTK_TREE_VIEW (app_search->tree_view),
                                           path, NULL, &rect);
        gtk_tree_view_convert_bin_window_to_tree_coords (GTK_TREE_VIEW (app_search->tree_view),
                                                         0, rect.y, NULL, &y);
        hildon_pannable_area_scroll_to (app_search->pannable_area, -1, y);
        gtk_tree_path_free (path);

        return TRUE;
    }

    if (widget == GTK_WIDGET(app_search->tree_view)) {
        GdkEvent *new_event;
        gboolean handled;

        if (!GTK_WIDGET_HAS_FOCUS(app_search->entry)) {
            gtk_widget_grab_focus(app_search->entry);
        }

        new_event = gdk_event_copy((GdkEvent*)event);
        handled = gtk_widget_event(app_search->entry, new_event);

        gdk_event_free(new_event);
        return handled;
    }

    return FALSE;
}

static void
_refilter_by_text(GtkEditable *entry,
		  gpointer     user_data)
{
    AppSearch *app_search = (AppSearch*)user_data;
    GtkTreeSelection *selection;
    GtkTreeIter iter;

    g_return_if_fail(GTK_IS_TREE_MODEL_FILTER(app_search->filtered_app_list));

    GtkTreeModelFilter* filter = GTK_TREE_MODEL_FILTER(app_search->filtered_app_list);
    g_object_ref(filter);
    gtk_tree_view_set_model(app_search->tree_view, NULL);
    gtk_tree_model_filter_refilter(filter);
    gtk_tree_view_set_model(app_search->tree_view, GTK_TREE_MODEL(filter));
    g_object_unref(filter);

    /*selection = gtk_tree_view_get_selection(app_search->tree_view);
    if (!gtk_tree_selection_get_selected(selection, NULL, &iter)) {
        if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(filter), &iter)) {
            gtk_tree_selection_select_iter(selection, &iter);
        }
    }*/
}

static void
show_app_search(AppSearch* app_search)
{
  GtkWidget* dialog;
  GtkWidget* search_bar;
  GtkWidget *app_selector;
  GtkTreeSelection* selection;
  gtk_list_store_clear(app_search->app_list);

  dialog = gtk_dialog_new();
  app_search->dialog = dialog;
  gtk_window_set_title(GTK_WINDOW(dialog), "");
  app_selector = hildon_gtk_tree_view_new(HILDON_UI_MODE_EDIT);
  app_search->tree_view = GTK_TREE_VIEW(app_selector);
  selection = gtk_tree_view_get_selection(app_search->tree_view);
  search_bar = hildon_entry_new(HILDON_SIZE_FINGER_HEIGHT);
  app_search->entry = search_bar;
  g_object_set(search_bar,"hildon-input-mode",
               HILDON_GTK_INPUT_MODE_FULL, NULL);


  _app_list_init_func(app_search);

  GtkTreeViewColumn* column_icon = gtk_tree_view_column_new();
  GtkTreeViewColumn* column_app_name = gtk_tree_view_column_new();
  GtkCellRenderer* icon_renderer = gtk_cell_renderer_pixbuf_new();
  GtkCellRenderer* app_name_renderer = gtk_cell_renderer_text_new();

  gtk_tree_view_append_column(GTK_TREE_VIEW(app_selector),
			      column_icon);
  gtk_tree_view_append_column(GTK_TREE_VIEW(app_selector),
			      column_app_name);
  gtk_tree_view_column_pack_start(column_icon, icon_renderer, FALSE);
  gtk_tree_view_column_pack_start(column_app_name, app_name_renderer, TRUE);  

  gtk_tree_view_column_set_expand(column_icon, FALSE);
  gtk_tree_view_column_set_expand(column_app_name, TRUE);

  g_object_set(icon_renderer, "stock-size", GTK_ICON_SIZE_BUTTON, NULL);
  g_object_set(app_name_renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL);

  gtk_tree_view_column_set_attributes(column_icon,
				      icon_renderer,
				      "icon-name",0,
				      NULL);

  gtk_tree_view_column_set_attributes(column_app_name,
				      app_name_renderer,
				      "markup",1,
				      NULL);

  gtk_tree_view_set_model(GTK_TREE_VIEW(app_selector),
			  GTK_TREE_MODEL(app_search->filtered_app_list));

  GtkWidget* pan = hildon_pannable_area_new();
  app_search->pannable_area = HILDON_PANNABLE_AREA(pan);
  gtk_container_add(GTK_CONTAINER(pan), app_selector);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), pan, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), search_bar,FALSE,FALSE,0); 
  g_signal_connect(search_bar, "key-press-event", G_CALLBACK(_on_key_press), app_search);
  g_signal_connect(app_selector, "key-press-event", G_CALLBACK(_on_key_press), app_search);
  g_signal_connect(search_bar, "changed", G_CALLBACK(_refilter_by_text), app_search);
  g_signal_connect(search_bar, "activate", G_CALLBACK(_on_activate), app_search);
  gtk_window_set_default_size(GTK_WINDOW(dialog), -1, 350);
  gtk_widget_grab_focus(search_bar);
  gtk_widget_show_all(dialog);
  gtk_dialog_run(GTK_DIALOG(dialog));
  gtk_widget_destroy(dialog);
}


void
app_search_init(AppSearch* app_search)
{
  app_search->app_list = gtk_list_store_new(5, G_TYPE_STRING, G_TYPE_STRING, 
						      G_TYPE_STRING, 
						      G_TYPE_STRING, 
						      G_TYPE_STRING);
  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(app_search->app_list),
				       1, GTK_SORT_ASCENDING);
  app_search->filtered_app_list = gtk_tree_model_filter_new(GTK_TREE_MODEL(app_search->app_list),
								      NULL);
  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(app_search->filtered_app_list),
					 _filter_app_list_func,
					 app_search,
					 NULL);
}


int main(int argc, char *argv[])
{
    AppSearch app_search;

    gtk_init(&argc, &argv);
    app_search_init(&app_search);

    show_app_search(&app_search);

    return 0;
}

